# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Processor",)

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from ._marker_relation import MarkerRelation
from ._materialization import Materialization
from ._operation_relations import BinaryOperationRelation, UnaryOperationRelation
from ._operations import Chain
from ._transfer import Transfer

if TYPE_CHECKING:
    from ._engine import Engine
    from ._relation import Relation


class Processor(ABC):
    """An inheritable framework for processing multi-engine relation trees.

    Individual `Engine` classes have different definitions of what it means to
    process a relation tree, and no single engine can handle a tree with
    engines.  This class provides a recursive algorithm that fills that role,
    with abstract method hooks for implementing `Transfer` and `Materialize`
    operations.

    Notes
    -----
    The `Processor` algorithm walks the tree recursively until it either finds:

    - a `Relation` with a `Relation.payload` that is not `None`, which is
      returned as-is;

    - a `Materialization` operation, for which a payload is computed via a call
      to the `materialize` hook, and then attached to both the original
      relation (modifying it in-place) and the returned one;

    - a `Transfer` operation, for which a payload is computed via a call to
      the `transfer` hook, and then the attached to the returned relation only.

    In addition, `Processor` never calls either hook on
    `trivial <Relation.is_trivial>` methods -
    `Engine.get_join_identity_payload` and `Engine.get_doomed_payload` are
    called instead.  This can (for example) avoid executing asking a database
    to execute a SQL query when the relation tree knows in advance the result
    will have no real content.  It also special-cases `Transfer` operations
    that are followed immediately by a `Materialization`, allowing both
    operations to be handled by a single call.
    """

    def process(self, relation: Relation) -> Relation:
        """Main entry point for processing a relation tree.

        Parameters
        ----------
        relation : `Relation`
            Root of the relation tree to process.  On return, relations that
            hold a `Materialization` relation will have a new
            `~Relation.payload` attached, if they did not have one already.

        Returns
        -------
        processed : `Relation`
            A version of the relation tree in which any relation with a
            `Transfer` operation has a copy of the original `Transfer` that
            has a `~Relation.payload` attached.
        """
        return self._process_recursive(relation, materialize_as=None)[0]

    @abstractmethod
    def transfer(self, source: Relation, destination: Engine, materialize_as: str | None) -> Any:
        """Hook for implementing transfers between engines.

        This method should be called only by the `Processor` base class.

        Parameters
        ----------
        source : `Relation`
            Relation to be transferred.  Any upstream `Transfer` operations in
            this tree are guaranteed to already have a `~Relation.payload`
            already attached (or some intervening relation does), so the
            relation's own engine should be capable of processing it on its
            own.
        destination : `Engine`
            Engine the relation is being transferred to.
        materialize_as : `str` or `None`
            If not `None`, the name of a `Materialization` operation that
            immediately follows the transfer being implemented, in which case
            the returned `~Relation.payload` should be appropriate for caching
            with the `Materialization`.

        Returns
        -------
        payload
            Payload for this relation in the ``destination`` engine.
        """
        raise NotImplementedError()

    @abstractmethod
    def materialize(self, target: Relation, name: str) -> Any:
        """Hook for implementing materialization operations.

        This method should be called only by the `Processor` base class.

        Parameters
        ----------
        target : `Relation`
            Relation to be materialized.  Any upstream `Transfer` operations in
            this tree are guaranteed to already have a `~Relation.payload`
            already attached (or some intervening relation does), so the
            relation's own engine should be capable of processing it on its
            own.
        name : `str`
            The name of the `Materialization` operation, to be used as needed
            in the engine-specific payload.

        Returns
        -------
        payload
            Payload for this relation that should be cached.
        """
        raise NotImplementedError()

    def _process_recursive(self, original: Relation, materialize_as: str | None) -> tuple[Relation, bool]:
        """Recursive implementation for `process`.

        Parameters
        ----------
        original : `Relation`
            Relation from the tree originally passed to `process`.
        materialize_as : `str` | `None`
            The name of the `Materialization` operation just downstream of this
            call, or `None` if the caller was not `_process_recursive` itself
            acting on a a `Materialization` operation.

        Returns
        -------
        processed : `Relation`
            Relation tree with `~Relation.payload` values attached to any
            `Transfer` operations.
        was_materialized : `bool`
            If `True`, `transfer` was called with ``materialize_as`` not
            `None`, and hence the caller (which must have been
            `_process_recursive` acting on a `Materialization` operation) does
            not need to call `materialize` to obtain a payload suitable for
            materialization.
        """
        if original.payload is not None:
            return original, True
        result: Relation
        payload: Any = None
        match original:
            case Transfer(destination=destination, target=target):
                # If the result is a trivial relation, just make a new
                # payload directly in the destination engine.
                if original.is_join_identity:
                    payload = destination.get_join_identity_payload()
                    new_target = target
                elif original.max_rows == 0:
                    payload = destination.get_doomed_payload(original.columns)
                    new_target = target
                else:
                    # Process recursively, ensuring upstream transfers
                    # and materializations happen first.
                    new_target, _ = self._process_recursive(target, materialize_as=None)
                    # Actually execute the transfer.  If materialize_as
                    # is not None, this will also take care of an
                    # immediately-downstream Materialization.
                    payload = self.transfer(new_target, destination, materialize_as)
                # We need to attach this payload to the processed
                # relation we return, but we don't want to attach it to
                # the original, so we reapply the transfer operation to
                # new_target even if new_target is target.
                result = original.reapply(new_target, payload)
                return result, materialize_as is not None
            case Materialization(name=name, target=target):
                assert name is not None, "Guaranteed by Materialization.apply."
                # Process recursively, ensuring upstream transfers and
                # materializations happen first.  Pass name as
                # materialize_as to tell an immediately-upstream
                # transfer to materialize directly.
                new_target, persisted = self._process_recursive(target, materialize_as=name)
                if new_target is not target:
                    result = new_target.materialized(name=name)
                    if result.payload is not None:
                        # This operation has been simplified away
                        # (perhaps it's now a materialization of a
                        # leaf).
                        original.attach_payload(result.payload)
                        return result, True
                else:
                    result = original
                if persisted:
                    payload = new_target.payload
                elif original.is_join_identity:
                    payload = target.engine.get_join_identity_payload()
                elif original.max_rows == 0:
                    payload = target.engine.get_doomed_payload(original.columns)
                else:
                    payload = self.materialize(new_target, name)
                # Attach the payload to the original relation, not just
                # the processed one, so it's used every time that the
                # original relation tree is processed.
                original.attach_payload(payload)
                if result is not original:
                    result.attach_payload(payload)
                return result, True
            case MarkerRelation(target=target):
                new_target, persisted = self._process_recursive(target, materialize_as=materialize_as)
                return original.reapply(new_target), persisted
            case UnaryOperationRelation(operation=operation, target=target):
                new_target, _ = self._process_recursive(target, materialize_as=None)
                if new_target is not target:
                    return operation.apply(new_target), False
                else:
                    return original, False
            case BinaryOperationRelation(operation=operation, lhs=lhs, rhs=rhs):
                new_lhs, lhs_persisted = self._process_recursive(lhs, materialize_as=None)
                new_rhs, rhs_persisted = self._process_recursive(rhs, materialize_as=None)
                if isinstance(operation, Chain):
                    # Simplify out relations with no rows from unions to save
                    # engines from having to handle those do-nothing branches.
                    # We don't do that earlier to the original tree usually
                    # because this is useful diagnostic information.
                    if new_lhs.max_rows == 0:
                        return new_rhs, rhs_persisted
                    if new_rhs.max_rows == 0:
                        return new_lhs, lhs_persisted
                if new_lhs is not lhs or new_rhs is not rhs:
                    return operation.apply(new_lhs, new_rhs), False
                return original, False
        raise AssertionError("Match should be exhaustive and all branches should return.")



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Deduplication",)

import dataclasses
from typing import TYPE_CHECKING, Literal, final

from .._operation_relations import UnaryOperationRelation
from .._unary_operation import UnaryCommutator, UnaryOperation

if TYPE_CHECKING:
    from .._relation import Relation


@final
@dataclasses.dataclass(frozen=True)
class Deduplication(UnaryOperation):
    """A relation operation that removes duplicate rows."""

    @property
    def is_count_invariant(self) -> Literal[False]:
        # Docstring inherited.
        return False

    @property
    def is_empty_invariant(self) -> Literal[True]:
        # Docstring inherited.
        return True

    def __str__(self) -> str:
        return "deduplicate"

    def applied_min_rows(self, target: Relation) -> int:
        # Docstring inherited.
        return 1

    def commute(self, current: UnaryOperationRelation) -> UnaryCommutator:
        # Docstring inherited.
        # Deduplication does not commute through Projection, but this is a bit
        # more defensive to guard against what a Projection does that's
        # problematic, rather than just checking isinstance(Projection).
        if not current.columns >= current.target.columns:
            return UnaryCommutator(
                first=None,
                second=current.operation,
                done=False,
                messages=(
                    "deduplication columns would change from "
                    f"{set(current.columns)} to {set(current.target.columns)}",
                ),
            )
        if current.operation.is_count_dependent:
            return UnaryCommutator(
                first=None,
                second=current.operation,
                done=False,
                messages=(f"{current.operation} is count-dependent",),
            )
        return UnaryCommutator(self, current.operation)



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import unittest
from collections.abc import Mapping

import sqlalchemy
from lsst.daf.relation import (
    ColumnContainer,
    ColumnExpression,
    ColumnFunction,
    ColumnTag,
    LogicalAnd,
    LogicalOr,
    Predicate,
    flatten_logical_and,
    iteration,
    sql,
    tests,
)


class ColumnExpressionTestCase(tests.RelationTestCase):
    def test_operator_function(self) -> None:
        """Test ColumnFunction via its constructor and a name found in the
        `operator` module.
        """
        tag = tests.ColumnTag("tag")
        expression = ColumnFunction(
            "__sub__",
            (ColumnExpression.reference(tag), ColumnExpression.literal(5)),
            dtype=int,
            supporting_engine_types=None,
        )
        self.assertEqual(expression.dtype, int)
        self.assertEqual(expression.supporting_engine_types, None)
        self.assertEqual(expression.args, (ColumnExpression.reference(tag), ColumnExpression.literal(5)))
        self.assertEqual(expression.name, "__sub__")
        self.assertEqual(expression.columns_required, {tag})
        self.assertEqual(
            str(expression),
            "__sub__(tag, 5)",
        )
        iteration_engine = iteration.Engine()
        self.assertTrue(expression.is_supported_by(iteration_engine))
        callable = iteration_engine.convert_column_expression(expression)
        self.assertEqual(callable({tag: 3}), -2)
        self.assertEqual(callable({tag: 6}), 1)
        sql_engine = sql.Engine[sqlalchemy.sql.ColumnElement]()
        self.assertTrue(expression.is_supported_by(sql_engine))
        sql_expression = sql_engine.convert_column_expression(
            expression, {tag: sqlalchemy.schema.Column("tag")}
        )
        self.check_sql_str("tag - 5", sql_expression)

    def test_method(self) -> None:
        """Test ColumnFunction via ColumnExpression.method and a name found
        on the object itself.
        """
        tag = tests.ColumnTag("tag")
        expression = ColumnExpression.reference(tag).method("lower")
        self.assertEqual(expression.dtype, None)
        self.assertEqual(expression.supporting_engine_types, None)
        self.assertEqual(expression.args, (ColumnExpression.reference(tag),))
        self.assertEqual(expression.name, "lower")
        self.assertEqual(expression.columns_required, {tag})
        engine = iteration.Engine()
        self.assertTrue(expression.is_supported_by(engine))
        callable = engine.convert_column_expression(expression)
        self.assertEqual(callable({tag: "MiXeDcAsE"}), "mixedcase")

    def test_engine_function(self) -> None:
        """Test ColumnFunction via ColumnExpression.function and a name
        that references a callable held by the engine.
        """
        tag = tests.ColumnTag("tag")
        engine = iteration.Engine()
        engine.functions["test_function"] = lambda x: x**2
        expression = ColumnExpression.function(
            "test_function",
            ColumnExpression.reference(tag),
            supporting_engine_types={iteration.Engine},
            dtype=int,
        )
        self.assertEqual(expression.dtype, int)
        self.assertEqual(expression.supporting_engine_types, (iteration.Engine,))
        self.assertEqual(expression.args, (ColumnExpression.reference(tag),))
        self.assertEqual(expression.name, "test_function")
        self.assertEqual(expression.columns_required, {tag})
        self.assertTrue(expression.is_supported_by(engine))
        self.assertFalse(expression.is_supported_by(sql.Engine[sqlalchemy.sql.ColumnElement]()))
        callable = engine.convert_column_expression(expression)
        self.assertEqual(callable({tag: 3}), 9)
        self.assertEqual(callable({tag: 4}), 16)

    def test_operator_predicate_function(self) -> None:
        """Test PredicateFunction via ColumnExpression factory methods and
        names found in the `operator` module.
        """
        tag = tests.ColumnTag("tag")
        ref = ColumnExpression.reference(tag, dtype=int)
        zero = ColumnExpression.literal(0, dtype=int)
        expressions = [ref.eq(zero), ref.ne(zero), ref.lt(zero), ref.le(zero), ref.gt(zero), ref.ge(zero)]
        self.assertEqual([x.dtype for x in expressions], [bool] * 6)
        self.assertEqual([x.supporting_engine_types for x in expressions], [None] * 6)
        self.assertEqual([x.args for x in expressions], [(ref, zero)] * 6)
        self.assertEqual(
            [x.name for x in expressions], ["__eq__", "__ne__", "__lt__", "__le__", "__gt__", "__ge__"]
        )
        self.assertEqual([x.columns_required for x in expressions], [{tag}] * 6)
        self.assertEqual(
            [str(x) for x in expressions],
            ["tag=0", "tag≠0", "tag<0", "tag≤0", "tag>0", "tag≥0"],
        )
        iteration_engine = iteration.Engine()
        self.assertEqual([x.is_supported_by(iteration_engine) for x in expressions], [True] * 6)
        callables = [iteration_engine.convert_predicate(x) for x in expressions]
        self.assertEqual([c({tag: 0}) for c in callables], [True, False, False, True, False, True])
        self.assertEqual([c({tag: 1}) for c in callables], [False, True, False, False, True, True])
        self.assertEqual([c({tag: -1}) for c in callables], [False, True, True, True, False, False])
        sql_bind = sqlalchemy.schema.Column("tag")
        sql_engine = sql.Engine[sqlalchemy.sql.ColumnElement]()
        self.assertEqual([x.is_supported_by(sql_engine) for x in expressions], [True] * 6)
        sql_expressions: list[sqlalchemy.sql.ColumnElement] = [
            sql_engine.convert_predicate(x, {tag: sql_bind}) for x in expressions
        ]
        self.assertEqual(
            [tests.to_sql_str(s) for s in sql_expressions],
            [
                "tag = 0",
                "tag != 0",
                "tag < 0",
                "tag <= 0",
                "tag > 0",
                "tag >= 0",
            ],
        )

    def test_column_expression_sequence(self) -> None:
        """Test ColumnExpressionSequence and ColumnInContainer."""
        a = tests.ColumnTag("a")
        seq = ColumnContainer.sequence(
            [ColumnExpression.reference(a, dtype=int), ColumnExpression.literal(4, dtype=int)], dtype=int
        )
        self.assertEqual(seq.dtype, int)
        self.assertEqual(
            list(seq.items),
            [ColumnExpression.reference(a, dtype=int), ColumnExpression.literal(4, dtype=int)],
        )
        self.assertEqual(seq.columns_required, {a})
        self.assertEqual(str(seq), "[a, 4]")
        iteration_engine = iteration.Engine()
        sql_engine = sql.Engine[sqlalchemy.sql.ColumnElement]()
        self.assertTrue(seq.is_supported_by(iteration_engine))
        self.assertTrue(seq.is_supported_by(sql_engine))
        b = tests.ColumnTag("b")
        contains = seq.contains(ColumnExpression.reference(b))
        self.assertEqual(contains.dtype, bool)
        self.assertEqual(contains.columns_required, {a, b})
        self.assertEqual(str(contains), "b∈[a, 4]")
        self.assertTrue(contains.is_supported_by(iteration_engine))
        self.assertTrue(contains.is_supported_by(sql_engine))
        callable = iteration_engine.convert_predicate(contains)
        self.assertEqual(callable({a: 0, b: 0}), True)
        self.assertEqual(callable({a: 0, b: 3}), False)
        self.assertEqual(callable({a: 0, b: 4}), True)
        bind_a = sqlalchemy.schema.Column("a")
        bind_b = sqlalchemy.schema.Column("b")
        self.check_sql_str("b IN (a, 4)", sql_engine.convert_predicate(contains, {a: bind_a, b: bind_b}))

    def test_range_literal(self) -> None:
        """Test ColumnRangeLiteral and ColumnInContainer."""
        ranges = [
            ColumnContainer.range_literal(range(3, 4)),
            ColumnContainer.range_literal(range(3, 6)),
            ColumnContainer.range_literal(range(2, 11, 3)),
        ]
        self.assertEqual([r.dtype for r in ranges], [int] * 3)
        self.assertEqual([r.columns_required for r in ranges], [frozenset()] * 3)
        self.assertEqual([str(r) for r in ranges], ["[3:4:1]", "[3:6:1]", "[2:11:3]"])
        a = tests.ColumnTag("a")
        contains = [r.contains(ColumnExpression.reference(a)) for r in ranges]
        self.assertEqual([c.dtype for c in contains], [bool] * 3)
        self.assertEqual([c.columns_required for c in contains], [{a}] * 3)
        self.assertEqual([str(c) for c in contains], ["a∈[3:4:1]", "a∈[3:6:1]", "a∈[2:11:3]"])
        iteration_engine = iteration.Engine()
        self.assertEqual([c.is_supported_by(iteration_engine) for c in contains], [True] * 3)
        callables = [iteration_engine.convert_predicate(c) for c in contains]
        self.assertEqual([c({a: 3}) for c in callables], [True, True, False])
        self.assertEqual([c({a: 5}) for c in callables], [False, True, True])
        self.assertEqual([c({a: 8}) for c in callables], [False, False, True])
        sql_engine = sql.Engine[sqlalchemy.sql.ColumnElement]()
        self.assertEqual([c.is_supported_by(sql_engine) for c in contains], [True] * 3)
        bind_a = sqlalchemy.schema.Column("a")
        sql_expressions = [sql_engine.convert_predicate(c, {a: bind_a}) for c in contains]
        self.assertEqual(
            [tests.to_sql_str(s) for s in sql_expressions],
            ["a = 3", "a BETWEEN 3 AND 5", "a BETWEEN 2 AND 10 AND a % 3 = 2"],
        )

    def test_logical_operators(self) -> None:
        """Test predicate logical operator expressions, Predicate.as_literal,
        and flatten_logical_and.
        """
        a = tests.ColumnTag("a")
        b = tests.ColumnTag("b")
        t = Predicate.literal(True)
        f = Predicate.literal(False)
        x = ColumnExpression.reference(a).gt(ColumnExpression.literal(0))
        y = Predicate.reference(b)
        iteration_engine = iteration.Engine()
        sql_engine = sql.Engine[sqlalchemy.sql.ColumnElement]()
        # Check attributes and simple accessors for predicate literals and
        # references.  Some as_trivial overloads are not checked because MyPy
        # can tell they always return None and complains if we try to use that
        # return value (which also means MyPy takes care of that "test" for us,
        # even though coverage can't tell).
        self.assertEqual(t.columns_required, frozenset())
        self.assertEqual(f.columns_required, frozenset())
        self.assertEqual(y.columns_required, {b})
        self.assertEqual(str(t), "True")
        self.assertEqual(str(f), "False")
        self.assertEqual(str(y), "b")
        self.assertIs(t.as_trivial(), True)
        self.assertIs(f.as_trivial(), False)
        self.assertTrue(t.is_supported_by(iteration_engine))
        self.assertTrue(f.is_supported_by(iteration_engine))
        self.assertTrue(y.is_supported_by(iteration_engine))
        self.assertTrue(t.is_supported_by(sql_engine))
        self.assertTrue(f.is_supported_by(sql_engine))
        self.assertTrue(y.is_supported_by(sql_engine))
        # Test factory methods for logical operators, including simplification.
        self.assertIs(t.logical_not().as_trivial(), False)
        self.assertIs(f.logical_not().as_trivial(), True)
        self.assertIs(t.logical_and(x).as_trivial(), None)
        self.assertIs(f.logical_and(x).as_trivial(), False)
        self.assertIs(t.logical_or(x).as_trivial(), True)
        self.assertIs(f.logical_or(x).as_trivial(), None)
        self.assertEqual(Predicate.logical_and(), t)
        self.assertEqual(Predicate.logical_or(), f)
        self.assertEqual(Predicate.logical_and(x), x)
        self.assertEqual(Predicate.logical_or(x), x)
        # Test attributes and simple accessors for logical operators.
        not_x = x.logical_not()
        self.assertEqual(not_x.columns_required, {a})
        self.assertEqual(str(not_x), "not (a>0)")
        self.assertIs(not_x.as_trivial(), None)
        self.assertTrue(not_x.is_supported_by(iteration_engine))
        self.assertTrue(not_x.is_supported_by(sql_engine))
        x_and_y = x.logical_and(y)
        self.assertEqual(x_and_y.columns_required, {a, b})
        self.assertEqual(str(x_and_y), "a>0 and b")
        self.assertIs(x_and_y.as_trivial(), None)
        self.assertTrue(x_and_y.is_supported_by(iteration_engine))
        self.assertTrue(x_and_y.is_supported_by(sql_engine))
        x_or_y = x.logical_or(y)
        self.assertEqual(x_or_y.columns_required, {a, b})
        self.assertEqual(str(x_or_y), "a>0 or b")
        self.assertIs(x_or_y.as_trivial(), None)
        self.assertTrue(x_or_y.is_supported_by(iteration_engine))
        self.assertTrue(x_or_y.is_supported_by(sql_engine))
        # Test iteration engine conversions.
        self.assertEqual(iteration_engine.convert_predicate(t)({}), True)
        self.assertEqual(iteration_engine.convert_predicate(f)({}), False)
        self.assertEqual(iteration_engine.convert_predicate(y)({b: True}), True)
        self.assertEqual(iteration_engine.convert_predicate(y)({b: False}), False)
        self.assertEqual(iteration_engine.convert_predicate(not_x)({a: 1}), False)
        self.assertEqual(iteration_engine.convert_predicate(not_x)({a: 0}), True)
        self.assertEqual(iteration_engine.convert_predicate(x_and_y)({a: 1, b: True}), True)
        self.assertEqual(iteration_engine.convert_predicate(x_and_y)({a: 1, b: False}), False)
        self.assertEqual(iteration_engine.convert_predicate(x_and_y)({a: 0, b: True}), False)
        self.assertEqual(iteration_engine.convert_predicate(x_and_y)({a: 0, b: False}), False)
        self.assertEqual(iteration_engine.convert_predicate(x_or_y)({a: 1, b: True}), True)
        self.assertEqual(iteration_engine.convert_predicate(x_or_y)({a: 1, b: False}), True)
        self.assertEqual(iteration_engine.convert_predicate(x_or_y)({a: 0, b: True}), True)
        self.assertEqual(iteration_engine.convert_predicate(x_or_y)({a: 0, b: False}), False)
        # Test SQL engine conversions.
        columns: Mapping[ColumnTag, sqlalchemy.sql.ColumnElement] = {
            a: sqlalchemy.schema.Column("a"),
            b: sqlalchemy.schema.Column("b"),
        }
        self.check_sql_str(
            "1",
            sql_engine.convert_predicate(t, columns),
            sql_engine.convert_predicate(LogicalAnd(()), columns),
        )
        self.check_sql_str(
            "0",
            sql_engine.convert_predicate(f, columns),
            sql_engine.convert_predicate(LogicalOr(()), columns),
        )
        self.check_sql_str(
            "b",
            sql_engine.convert_predicate(y, columns),
            sql_engine.convert_predicate(LogicalAnd((y,)), columns),
            sql_engine.convert_predicate(LogicalOr((y,)), columns),
        )
        # Apparently SQLAlchemy does some simplifications of its own on NOT
        # operations.
        self.check_sql_str("a <= 0", sql_engine.convert_predicate(not_x, columns))
        self.check_sql_str("a > 0 AND b", sql_engine.convert_predicate(x_and_y, columns))
        self.check_sql_str("a > 0 OR b", sql_engine.convert_predicate(x_or_y, columns))
        # Test flatten_logical_and
        self.assertEqual(flatten_logical_and(t), [])
        self.assertIs(flatten_logical_and(f), False)
        self.assertEqual(flatten_logical_and(not_x), [not_x])
        self.assertEqual(flatten_logical_and(x_and_y), [x, y])
        self.assertEqual(flatten_logical_and(x_or_y), [x_or_y])
        self.assertEqual(flatten_logical_and(x_and_y.logical_and(t)), [x, y])
        self.assertEqual(flatten_logical_and(x_and_y.logical_and(f)), False)
        c = tests.ColumnTag("c")
        z = Predicate.reference(c)
        self.assertEqual(flatten_logical_and(x_and_y.logical_and(z)), [x, y, z])


if __name__ == "__main__":
    unittest.main()



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("UnaryOperation", "RowFilter", "Reordering", "Identity", "UnaryCommutator")

import dataclasses
from abc import ABC, abstractmethod
from collections.abc import Set
from typing import TYPE_CHECKING, Literal, final

from ._columns import ColumnTag
from ._exceptions import EngineError
from ._relation import Relation

if TYPE_CHECKING:
    from ._engine import Engine
    from ._operation_relations import UnaryOperationRelation


class UnaryOperation(ABC):
    """An abstract base class for operations that act on a single relation.

    Notes
    -----
    A `UnaryOperation` represents the operation itself; the combination of an
    operation and the "target" relation it acts on to form a new relation is
    represented by the `UnaryOperationRelation` class.  That combination should
    always be created via a call to the `apply` method (or something that calls
    it, like the convenience methods on the `Relation` class).  In some cases,
    applying a `UnaryOperation` doesn't return something involving the original
    operation, because of some combination of defaulted-parameter population
    and simplification, and there are even some `UnaryOperation` classes that
    should never actually appear in a `UnaryOperationRelation`.

    `UnaryOperation` cannot be subclassed directly by external code, but it has
    two more restricted subclasses that can be:`RowFilter` and `Reordering`.

    All concrete `UnaryOperation` types are frozen, equality-comparable
    `dataclasses`.  They also provide a very concise `str` representation (in
    addition to the dataclass-provided `repr`) suitable for summarizing an
    entire relation tree.

    See Also
    --------
    :ref:`lsst.daf.relation-overview-operations`
    """

    def __init_subclass__(cls) -> None:
        assert (
            cls.__name__
            in {
                "Calculation",
                "Deduplication",
                "Identity",
                "PartialJoin",
                "Projection",
                "RowFilter",
                "Reordering",
            }
            or cls.__base__ is not UnaryOperation
        ), (
            "UnaryOperation inheritance is closed to predefined types in daf_relation, "
            "except for subclasses of RowFilter and Reordering."
        )

    @property
    def columns_required(self) -> Set[ColumnTag]:
        """The columns the target relation must have in order for this
        operation to be applied to it (`~collections.abc.Set` [ `ColumnTag` ]
        ).
        """
        return frozenset()

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_empty_invariant(self) -> bool:
        """Whether this operation can remove all rows from its target relation
        (`bool`).
        """
        raise NotImplementedError()

    @property
    @abstractmethod
    def is_count_invariant(self) -> bool:
        """Whether this operation can change the number of rows in its target
        relation (`bool`).

        The number of rows here includes duplicates - removing duplicates is
        not considered a count-invariant operation.
        """
        raise NotImplementedError()

    @property
    def is_order_dependent(self) -> bool:
        """Whether this operation depends on the order of the rows in its
        target relation (`bool`).
        """
        return False

    @property
    def is_count_dependent(self) -> bool:
        """Whether this operation depends on the number of rows in its target
        relation (`bool`).
        """
        return False

    def is_supported_by(self, engine: Engine) -> bool:
        """Whether this operation is supported by the given engine (`bool`)."""
        return True

    @final
    def apply(
        self,
        target: Relation,
        *,
        preferred_engine: Engine | None = None,
        backtrack: bool = True,
        transfer: bool = False,
        require_preferred_engine: bool = False,
    ) -> Relation:
        """Create a new relation that represents the action of this operation
        on an existing relation.

        Parameters
        ----------
        target : `Relation`
            Relation the operation will act on.
        preferred_engine : `Engine`, optional
            Engine that the operation would ideally be performed in.  If this
            is not equal to ``target.engine``, the ``backtrack``, ``transfer``,
            and ``require_preferred_engine`` arguments control the behavior.
            Some operations may supply their own preferred engine default, such
            as the "fixed" operand's own engine in a `PartialJoin`.
        backtrack : `bool`, optional
            If `True` (default) and the current engine is not the preferred
            engine, attempt to insert this operation before a transfer upstream
            of the current relation, as long as this can be done without
            breaking up any locked relations or changing the resulting relation
            content.
        transfer : `bool`, optional
            If `True` (`False` is default) and the current engine is not the
            preferred engine, insert a new `Transfer` to the preferred engine
            before this operation.  If ``backtrack`` is also true, the transfer
            is added only if the backtrack attempt fails.
        require_preferred_engine : `bool`, optional
            If `True` (`False` is default) and the current engine is not the
            preferred engine, raise `EngineError`.  If ``backtrack`` is also
            true, the exception is only raised if the backtrack attempt fails.
            Ignored if ``transfer`` is true.

        Returns
        -------
        new_relation : `Relation`
            Relation that includes this operation.  This may be ``target`` if
            the operation is a no-op, and it may not be a
            `UnaryOperationRelation` holding this operation (or even a similar
            one) if the operation was inserted earlier in the tree via
            commutation relations or if simplification occurred.

        Raises
        ------
        ColumnError
            Raised if the operation could not be applied due to problems with
            the target relation's columns.
        EngineError
            Raised if the operation could not be applied due to problems with
            the target relation's engine.

        Notes
        -----
        Adding operations to relation trees is a potentially complex process in
        order to give both the operation type and the engine to customize the
        opportunity to enforce their own invariants.  This `~typing.final`
        method provides the bulk of the high-level implementation, and is
        called by the `Relation` class's convenience methods with essentially
        no additional logic.  The overall sequence is as follows:

        - `apply` starts by delegating to `_begin_apply`, which
          allows operation classes to replace the operation object itself,
          perform initial checks, and set the preferred engine.
        - `apply` then performs the ``preferred_engine`` logic indicated by the
          ``backtrack`` ``transfer``, and ``require_preferred_engine`` options,
          delegating backtracking to `Engine.backtrack_unary`.
        - `Engine.backtrack_unary` will typically call back to `commute` to
          determine how and whether to move the new operation upstream of
          existing ones.
        - If backtracking is not used or is not fully successful, `apply` then
          delegates to `Engine.append_unary` to add the operation to the root
          of the relation tree.
        - The `Engine` methods are expected to delegate back to
          `_finish_apply` when they have identified the location in the tree
          where the new operation should be inserted.
        - `_finish_apply` is responsible for actually constructing the
          `UnaryOperationRelation` when appropriate.  The default
          implementation of `_finish_apply` also calls `simplify` to see if it
          is possible to merge the new operation with those immediately
          upstream of it or elide it entirely.
        """
        operation, preferred_engine = self._begin_apply(target, preferred_engine)
        done = False
        result = target
        if preferred_engine != target.engine:
            if backtrack:
                result, done, messages = target.engine.backtrack_unary(operation, target, preferred_engine)
            else:
                messages = ("backtracking insertion not requested by caller",)
            if not done:
                if transfer:
                    result = result.transferred_to(preferred_engine)
                elif require_preferred_engine:
                    raise EngineError(
                        f"No way to apply {operation} to {target} "
                        f"with required engine '{preferred_engine}': {'; '.join(messages)}."
                    )
        if not done:
            result = result.engine.append_unary(operation, result)
        return result

    def _begin_apply(
        self, target: Relation, preferred_engine: Engine | None
    ) -> tuple[UnaryOperation, Engine]:
        """A customization hook for the beginning of operation application.

        Parameter
        ---------
        target : `Relation`
            Relation the operation should act on, at least conceptually.  Later
            logic may actually apply the operation upstream of this relation,
            but only when the result of doing so would be equivalent to
            applying it here.
        preferred_engine : `Engine` or `None`
            Preferred engine passed to `apply`.

        Returns
        -------
        operation : `UnaryOperation`
            The operation to actually apply.  The default implementation
            returns ``self``.
        preferred_engine : `Engine`
            The engine to actually prefer.  The default implementation returns
            the given ``preferred_engine`` if it is not `None`, and
            ``target.engine`` if it is `None`.

        Notes
        -----
        This method provides an opportunity for operations to establish any
        invariants that must be satisfied only when the operation is part of
        a relation.  Implementations can also return an `Identity` instance
        when they can determine that the operation will do nothing when applied
        to the given target.
        """
        return self, preferred_engine if preferred_engine is not None else target.engine

    def _finish_apply(self, target: Relation) -> Relation:
        """A customization hook for the end of operation application.

        Parameters
        ----------
        target : `Relation`
            Relation the operation will act upon directly.

        Returns
        -------
        applied : `Relation`
            Result of applying this operation to the given target.  Usually -
            but not always - a `UnaryOperationRelation` that holds ``self`` and
            ``target``.

        Notes
        -----
        This method provides an opportunity for operations to perform final
        simplification at the point of insertion (which the default
        implementation does, via calls to `simplify`) and change the kind of
        relation produced (the default implementation constructs a
        `UnaryOperationRelation`).
        """
        from ._operation_relations import UnaryOperationRelation

        match target:
            case UnaryOperationRelation():
                if simplified := self.simplify(target.operation):
                    if simplified is target.operation:
                        return target
                    else:
                        return simplified._finish_apply(target.target)

        if not self.is_supported_by(target.engine):
            raise EngineError(f"Operation {self} is not supported by engine {target.engine}.")

        return UnaryOperationRelation(
            operation=self,
            target=target,
            columns=self.applied_columns(target),
        )

    def applied_columns(self, target: Relation) -> Set[ColumnTag]:
        """Return the columns of the relation that results from applying this
        operation to the given target.

        Parameters
        ----------
        target : `Relation`
            Relation the operation will act on.

        Returns
        -------
        columns : `~collections.abc.Set` [ `ColumnTag` ]
            Columns the new relation would have.
        """
        return target.columns

    @abstractmethod
    def applied_min_rows(self, target: Relation) -> int:
        """Return the minimum number of rows of the relation that results from
        applying this operation to the given target.

        Parameters
        ----------
        target : `Relation`
            Relation the operation will act on.

        Returns
        -------
        min_rows : `int`
            Minimum number of rows the new relation would have.
        """
        raise NotImplementedError()

    def applied_max_rows(self, target: Relation) -> int | None:
        """Return the maximum number of rows of the relation that results from
        applying this operation to the given target.

        Parameters
        ----------
        target : `Relation`
            Relation the operation will act on.

        Returns
        -------
        max_rows : `int` or `None`
            Maximum number of rows the new relation would have.
        """
        return target.max_rows

    def commute(self, current: UnaryOperationRelation) -> UnaryCommutator:
        """Describe whether and how this operation can be moved upstream of an
        existing one without changing the content of the resulting relation.

        Parameters
        ----------
        current : `UnaryOperationRelation`
            A unary operation relation that is the current logical target of
            ``self``.

        Returns
        -------
        commutator : `UnaryCommutator`
            A struct that either provides a version of ``current.operation``
            that can be applied to ``current.target`` after a possibly-modified
            version of ``self``, or an explanation of why this is impossible.

        Notes
        -----
        The `commute` implementations for the provided concrete
        `UnaryOperation` types assume that all unary operations preserve row
        order.  If this is not the case in an engine, that engine should not
        implement `Engine.backtrack_unary` or take this into account itself
        when determining whether operations commute.
        """
        return UnaryCommutator(
            first=None,
            second=current.operation,
            done=False,
            messages=(f"{self} does not commute with anything",),
        )

    def simplify(self, upstream: UnaryOperation) -> UnaryOperation | None:
        """Return a simplified combination of this operation with another.

        Parameters
        ----------
        upstream : `UnaryOperation`
            Operation that acts immediately prior to ``self``.

        Returns
        -------
        simplified : `UnaryOperation`
            Operation that combines the action of ``upstream`` followed by
            ``self``, or `None` if no such combination is possible.
        """
        return None


class RowFilter(UnaryOperation):
    """An extensible `UnaryOperation` subclass for operations that only remove
    rows from their target.
    """

    @final
    @property
    def is_count_invariant(self, engine: Engine | None = None) -> Literal[False]:
        # Docstring inherited.
        return False

    @property
    @abstractmethod
    def is_order_dependent(self) -> bool:
        # Docstring inherited.
        raise NotImplementedError()

    @final
    def applied_columns(self, target: Relation) -> Set[ColumnTag]:
        # Docstring inherited.
        return target.columns

    def applied_min_rows(self, target: Relation) -> int:
        # Docstring inherited.
        if target.min_rows == 0 or self.is_count_invariant:
            return target.min_rows
        elif self.is_empty_invariant:
            return 1
        else:
            return 0


@final
class Identity(UnaryOperation):
    """A concrete unary operation that does nothing.

    `Identity` operations never appear in relation trees; their `apply` method
    always just returns the target relation.
    """

    def __str__(self) -> str:
        return "identity"

    def _finish_apply(self, target: Relation) -> Relation:
        # Docstring inherited.
        return target

    @property
    def is_count_invariant(self) -> Literal[True]:
        # Docstring inherited.
        return True

    @property
    def is_empty_invariant(self) -> Literal[True]:
        # Docstring inherited.
        return True

    def applied_columns(self, target: Relation) -> Set[ColumnTag]:
        # Docstring inherited.
        return target.columns

    def applied_min_rows(self, target: Relation) -> int:
        # Docstring inherited.
        return target.min_rows

    def applied_max_rows(self, target: Relation) -> int | None:
        # Docstring inherited.
        return target.max_rows

    def commute(self, current: UnaryOperationRelation) -> UnaryCommutator:
        return UnaryCommutator(first=self, second=current.operation)

    def simplify(self, current: UnaryOperation) -> UnaryOperation:
        return current


class Reordering(UnaryOperation):
    """An extensible `UnaryOperation` subclass for operations that only reorder
    rows.
    """

    @final
    @property
    def is_count_invariant(self) -> Literal[True]:
        # Docstring inherited.
        return True

    @final
    @property
    def is_empty_invariant(self) -> Literal[True]:
        # Docstring inherited.
        return True

    @final
    def applied_columns(self, target: Relation) -> Set[ColumnTag]:
        # Docstring inherited.
        return target.columns

    @final
    def applied_min_rows(self, target: Relation) -> int:
        # Docstring inherited.
        return target.min_rows

    @final
    def applied_max_rows(self, target: Relation) -> int | None:
        # Docstring inherited.
        return target.max_rows


@dataclasses.dataclass
class UnaryCommutator:
    """A struct for the return value of `UnaryOperation.commute`."""

    first: UnaryOperation | None
    """The first operation to apply in the commuted sequence (`UnaryOperation`
    or `None`).

    When at least some commutation is possible, this is a possibly-modified
    version of ``current.operation``, where ``current`` is the argument to
    `UnaryOperation.commute`.  When it is `None`, either the commutation failed
    or the original operation will simplify away entirely (as indicated by
    ``done``).
    """

    second: UnaryOperation
    """The second operation to apply in the commuted sequence
    (`UnaryOperation`).

    When commutation is successful, this is usually ``self`` or a modification
    thereof.  When commutation is unsuccessful, this should be exactly
    ``current.operation``, where ``current`` is the argument to
    `UnaryOperation.commute`.
    """

    done: bool = True
    """Whether the commutation was fully successful (`bool`).

    When `False`, the original downstream relation (``self`` in call to
    `commute`) must still be applied after `first` (if not `None`) and
    `second`.  While `first` is usually `None` in this case, `Projection`
    operations (and possibily some extension operations) can be partially
    commuted
    """

    messages: tuple[str, ...] = dataclasses.field(default_factory=tuple)
    """Messages that describe why commutation either failed or only
    partially succeeded.
    """



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Engine",)

import dataclasses
import itertools
from collections.abc import Callable, Container, Mapping, Sequence, Set
from operator import attrgetter, itemgetter
from typing import Any

from .._columns import (
    ColumnContainer,
    ColumnExpression,
    ColumnExpressionSequence,
    ColumnFunction,
    ColumnInContainer,
    ColumnLiteral,
    ColumnRangeLiteral,
    ColumnReference,
    ColumnTag,
    LogicalAnd,
    LogicalNot,
    LogicalOr,
    Predicate,
    PredicateFunction,
    PredicateLiteral,
    PredicateReference,
)
from .._engine import Engine as BaseEngine
from .._engine import GenericConcreteEngine
from .._exceptions import EngineError
from .._leaf_relation import LeafRelation
from .._marker_relation import MarkerRelation
from .._materialization import Materialization
from .._operation_relations import BinaryOperationRelation, UnaryOperationRelation
from .._operations import Calculation, Chain, Deduplication, Join, Projection, Selection, Slice, Sort
from .._relation import Relation
from .._transfer import Transfer
from .._unary_operation import UnaryOperation
from ._row_iterable import (
    CalculationRowIterable,
    ChainRowIterable,
    MaterializedRowIterable,
    ProjectionRowIterable,
    RowIterable,
    RowMapping,
    RowSequence,
    SelectionRowIterable,
)


@dataclasses.dataclass(repr=False, eq=False, kw_only=True)
class Engine(GenericConcreteEngine[Callable[..., Any]]):
    """A concrete engine that treats relations as iterables with
    `~collections.abc.Mapping` rows.

    See the `.iteration` module documentation for details.
    """

    name: str = "iteration"

    def __repr__(self) -> str:
        return f"lsst.daf.relation.iteration.Engine({self.name!r})@{id(self):0x}"

    def make_leaf(
        self,
        columns: Set[ColumnTag],
        payload: MaterializedRowIterable,
        *,
        name: str = "",
        messages: Sequence[str] = (),
        name_prefix: str = "leaf",
        parameters: Any = None,
    ) -> LeafRelation:
        """Create a nontrivial leaf relation in this engine.

        This is a convenience method that simply forwards all arguments to
        the `.LeafRelation` constructor; see that class for details.
        """
        return LeafRelation(
            self,
            frozenset(columns),
            payload,
            min_rows=len(payload),
            max_rows=len(payload),
            messages=messages,
            name=name,
            name_prefix=name_prefix,
            parameters=parameters,
        )

    def get_join_identity_payload(self) -> RowIterable:
        # Docstring inherited.
        return RowMapping((), {(): {}})

    def get_doomed_payload(self, columns: Set[ColumnTag]) -> RowIterable:
        # Docstring inherited.
        return RowMapping((), {})

    def backtrack_unary(
        self, operation: UnaryOperation, tree: Relation, preferred: BaseEngine
    ) -> tuple[Relation, bool, tuple[str, ...]]:
        # Docstring inherited.
        if tree.is_locked:
            return tree, False, (f"{tree} is locked",)
        match tree:
            case UnaryOperationRelation(target=target):
                commutator = operation.commute(tree)
                if commutator.first is None:
                    return tree, commutator.done, commutator.messages
                else:
                    upstream, done, messages = self.backtrack_unary(commutator.first, target, preferred)
                    if upstream is not target:
                        result = commutator.second._finish_apply(upstream)
                    else:
                        result = tree
                    return (
                        result,
                        done and commutator.done,
                        commutator.messages + messages,
                    )
            case BinaryOperationRelation():
                return tree, False, ("backtracking through binary operations is not implemented",)
            case Transfer(target=target) as transfer:
                if target.engine == preferred:
                    return transfer.reapply(operation.apply(target)), True, ()
                else:
                    upstream, done, messages = target.engine.backtrack_unary(operation, target, preferred)
                    return (transfer.reapply(upstream), done, messages)
        raise NotImplementedError(f"Unsupported relation type {tree} for engine {self}.")

    def execute(self, relation: Relation) -> RowIterable:
        """Execute a native iteration relation, returning a Python iterable.

        Parameters
        ----------
        relation : `.Relation`
            Relation to execute.

        Returns
        -------
        rows : `RowIterable`
            Iterable over rows, with each row a mapping keyed by `.ColumnTag`.

        Notes
        -----
        This method does not in general iterate over the relation's rows; while
        some operations like `Sort` and `Deduplication` do require processing
        all rows up front (which will happen during a call to `execute`), most
        return lazy iterables that do little or nothing until actually iterated
        over.

        This method requires all relations in the tree to have the same engine
        (``self``).  Use the `.Processor` class to handle trees with multiple
        engines.
        """
        if relation.engine != self:
            raise EngineError(
                f"Engine {self!r} cannot operate on relation {relation} with engine {relation.engine!r}. "
                "Use lsst.daf.relation.process to evaluate transfers first."
            )
        if relation.max_rows == 0:
            return RowSequence([])
        if relation.is_join_identity:
            return RowSequence([{}])
        if (result := relation.payload) is not None:
            return result
        match relation:
            case UnaryOperationRelation(operation=operation, target=target):
                target_rows = self.execute(target)
                match operation:
                    case Calculation(tag=tag, expression=expression):
                        return CalculationRowIterable(
                            target_rows, tag, self.convert_column_expression(expression)
                        )
                    case Deduplication():
                        unique_key = tuple(tag for tag in relation.columns if tag.is_key)
                        return target_rows.to_mapping(unique_key)
                    case Projection(columns=columns):
                        return ProjectionRowIterable(target_rows, columns)
                    case Selection(predicate=predicate):
                        return SelectionRowIterable(target_rows, self.convert_predicate(predicate))
                    case Slice(start=start, stop=stop):
                        return target_rows.sliced(start, stop)
                    case Sort(terms=terms):
                        rows_list = list(target_rows)
                        # Python's built-in sorting methods are stable, but
                        # they don't provide a way to sort some keys in
                        # ascending order and others in descending order.  So
                        # we split the sequence of order-by terms into groups
                        # of consecutive terms with the same
                        # ascending/descending state.  At the same time, we
                        # use the visitor to transform each expression into a
                        # callable we can apply to each row.
                        grouped_by_ascending = [
                            (ascending, [self.convert_column_expression(t.expression) for t in terms])
                            for ascending, terms in itertools.groupby(terms, key=attrgetter("ascending"))
                        ]
                        # Now we can sort the full list of rows once per group,
                        # in reverse so sort terms at the start of the order-by
                        # sequence "win".
                        for ascending, callables in grouped_by_ascending[::-1]:
                            rows_list.sort(
                                key=lambda row: tuple(c(row) for c in callables),
                                reverse=not ascending,
                            )
                        return RowSequence(rows_list)
                    case _:
                        return self.apply_custom_unary_operation(operation, target)
            case BinaryOperationRelation(operation=operation, lhs=lhs, rhs=rhs):
                match operation:
                    case Chain():
                        return ChainRowIterable([self.execute(lhs), self.execute(rhs)])
                    case Join():
                        raise EngineError("Joins are not supported by the iteration engine.")
                raise EngineError(f"Custom binary operation {operation} is not supported.")
            case Materialization(target=target):
                result = self.execute(target).materialized()
                relation.attach_payload(result)
                return result
            case Transfer(destination=destination, target=target):
                if isinstance(target.engine, Engine):
                    # This is a transfer from another iteration engine
                    # (maybe a subclass).  We can handle that without
                    # requiring a Processor, and that's useful for at
                    # least unit testing (though maybe not anything
                    # else; it's not clear why you'd have a transfer
                    # between iteration engines otherwise).
                    return target.engine.execute(target)
                raise EngineError(
                    f"Engine {self!r} cannot handle transfer from "
                    f"{target.engine!r} to {destination!r}; "
                    "use `lsst.daf.relation.Processor` first to handle this operation."
                )
            case MarkerRelation(target=target):
                return self.execute(target)
        raise AssertionError("matches should be exhaustive and all branches should return")

    def convert_column_expression(
        self, expression: ColumnExpression
    ) -> Callable[[Mapping[ColumnTag, Any]], Any]:
        """Convert a `.ColumnExpression` to a Python callable.

        Parameters
        ----------
        expression : `.ColumnExpression`
            Expression to convert.

        Returns
        -------
        callable
            Callable that takes a single `~collections.abc.Mapping` argument
            (with `.ColumnTag` keys and regular Python values, representing a
            row in a relation), returning the evaluated expression as another
            regular Python value.
        """
        match expression:
            case ColumnLiteral(value=value):
                return lambda row: value
            case ColumnReference(tag=tag):
                return itemgetter(tag)
            case ColumnFunction(name=name, args=args):
                function = self.get_function(name)
                if function is not None:
                    arg_callables = [self.convert_column_expression(arg) for arg in args]
                    # MyPy doesn't see 'function' as not-None for some reason.
                    return lambda row: function(*[c(row) for c in arg_callables])  # type: ignore
                first, *rest = [self.convert_column_expression(arg) for arg in args]
                return lambda row: getattr(first(row), name)(*[r(row) for r in rest])
        raise AssertionError("matches should be exhaustive and all branches should return")

    def convert_column_container(
        self, expression: ColumnContainer
    ) -> Callable[[Mapping[ColumnTag, Any]], Container]:
        """Convert a `.ColumnContainer` to a Python callable.

        Parameters
        ----------
        expression : `.ColumnContainer`
            Expression to convert.

        Returns
        -------
        callable
            Callable that takes a single `~collections.abc.Mapping` argument
            (with `.ColumnTag` keys and regular Python values, representing a
            row in a relation), returning the evaluated expression as
            `collections.abc.Container` instance.
        """
        match expression:
            case ColumnRangeLiteral(value=value):
                return lambda row: value
            case ColumnExpressionSequence(items=items):
                item_callables = [self.convert_column_expression(item) for item in items]
                return lambda row: {c(row) for c in item_callables}
        raise AssertionError("matches should be exhaustive and all branches should return")

    def convert_predicate(self, predicate: Predicate) -> Callable[[Mapping[ColumnTag, Any]], bool]:
        """Convert a `Predicate` to a Python callable.

        Parameters
        ----------
        predicate : `Predicate`
            Expression to convert.

        Returns
        -------
        callable
            Callable that takes a single `~collections.abc.Mapping` argument
            (with `.ColumnTag` keys and regular Python values, representing a
            row in a relation), returning the evaluated expression as a `bool`.
        """
        match predicate:
            case PredicateFunction(name=name, args=args):
                function = self.get_function(name)
                if function is not None:
                    arg_callables = [self.convert_column_expression(arg) for arg in args]
                    # MyPy doesn't see 'function' as not-None in the capture
                    # for some reason.
                    return lambda row: function(*[c(row) for c in arg_callables])  # type: ignore
                first, *rest = [self.convert_column_expression(arg) for arg in args]
                return lambda row: getattr(first(row), name)(*[r(row) for r in rest])
            case LogicalAnd(operands=operands):
                operand_callables = [self.convert_predicate(arg) for arg in operands]
                return lambda row: all(c(row) for c in operand_callables)
            case LogicalOr(operands=operands):
                operand_callables = [self.convert_predicate(arg) for arg in operands]
                return lambda row: any(c(row) for c in operand_callables)
            case LogicalNot(operand=operand):
                target_callable = self.convert_predicate(operand)
                return lambda row: not target_callable(row)
            case PredicateReference(tag=tag):
                return itemgetter(tag)
            case PredicateLiteral(value=value):
                return lambda row: value
            case ColumnInContainer(item=item, container=container):
                item_callable = self.convert_column_expression(item)
                container_callable = self.convert_column_container(container)
                return lambda row: item_callable(row) in container_callable(row)
        raise AssertionError("matches should be exhaustive and all branches should return")

    def apply_custom_unary_operation(self, operation: UnaryOperation, target: Relation) -> RowIterable:
        """Convert a custom `.UnaryOperation` to a `RowIterable`.

        This method must be implemented in a subclass engine in order to
        support any custom `.UnaryOperation`.

        Parameters
        ----------
        operation : `.UnaryOperation`
            Operation to apply.  Guaranteed to be a `.Marker`, `.Reordering`,
            or `.RowFilter` subclass.
        target : `.Relation`
            Target of the unary operation.  Typically this will be passed to
            `execute` and the result used to construct a new `RowIterable`.

        Returns
        -------
        rows : `RowIterable`
            Iterable over rows, with each row a mapping keyed by `.ColumnTag`.
        """
        raise EngineError(f"Custom operation {operation} not supported by engine {self}.")
